import builtins
import dataclasses
import os
import os.path
import re
import typing

@dataclasses.dataclass
class DiffResults:

    more:list[str] = dataclasses.field(default_factory=lambda: [])
    less:list[str] = dataclasses.field(default_factory=lambda: [])
    diff:list[str] = dataclasses.field(default_factory=lambda: [])

def _fr(fn:str):

    with open(fn,'rb') as f:

        return f.read()

def get(dirs    :typing.Iterable[str]       =list(),
        filter  :typing.Callable[[str],bool]=lambda fn: True):

    results_list:list[tuple[str,str,DiffResults]] = []
    files_map        = {d:set(fn for dp,dns_,fns in os.walk(d) for fn in builtins.filter(filter, map(lambda fn: os.path.relpath(path=os.path.join(dp,fn),start=d), fns))) for d in set(dirs)}
    for d1,d2 in set(tuple(sorted((d1,d2))) for d1 in files_map for d2 in files_map if d1 != d2):

        results = DiffResults()
        results_list.append((d1,d2,results))
        results.more.extend(sorted(set(fn for fn in files_map[d1] if fn not in files_map[d2])))
        results.less.extend(sorted(set(fn for fn in files_map[d2] if fn not in files_map[d1])))
        results.diff.extend(sorted(set(fn for fn in files_map[d1] if fn not in results.more and _fr(os.path.join(d1,fn)) != _fr(os.path.join(d2,fn)))))
    
    return results_list

def main():

    import argparse
    
    class A:

        DIRECTORIES     = 'dirs'
        FILE_NAME_REGEX = 'fre'
        DUMP_FILE_NAME  = 'o'

    p = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                description    ='Determine differences between 2 or more directories')
    p.add_argument(f'{A.DIRECTORIES}',
                   help ='directories to analyse',
                   nargs='+')
    p.add_argument(f'--{A.FILE_NAME_REGEX}',
                   help   ='regular expression to filter files to compare\nDefaults to wildcard (admit all files).',
                   default='.*')
    p.add_argument(f'--{A.DUMP_FILE_NAME}',
                   help='name of file to which to dump the analysis result, if desired')
    get_ = p.parse_args().__getattribute__
    # do it
    dump_fn         = get_(A.DUMP_FILE_NAME) if get_(A.DUMP_FILE_NAME) else None
    f:typing.TextIO = None if dump_fn is None else open(dump_fn, mode='w')
    _dump :typing.Callable[[str],None] = (lambda s: None) if dump_fn is None else (lambda s: f.write(s+'\n'))
    _close:typing.Callable[[],None]    = (lambda:   None) if dump_fn is None else (lambda:   f.close())
    def dump(s:str):

        print(s)
        _dump(s)
    
    def close(): _close()
    results = get(dirs   =get_(A.DIRECTORIES),
                  filter =lambda fn,_p=re.compile(get_(A.FILE_NAME_REGEX)): _p.search(fn))
    for d1,d2,r in results:

        dump(f'Between {d1} and {d2}')
        if r.more: dump(f'\n  Files in {d1} not in {d2} ({len(r.more)}):\n{'\n'.join(f'  - {fn}' for fn in r.more)}')
        if r.less: dump(f'\n  Files in {d2} not in {d1} ({len(r.less)}):\n{'\n'.join(f'  - {fn}' for fn in r.less)}')
        if r.diff: dump(f'\n  Files differ between {d1} and {d2} ({len(r.diff)}):\n{'\n'.join(f'  - {fn}' for fn in r.diff)}')
    
    close()

if __name__ == '__main__': main()



def main():

    import argparse

    class A:

        FILENAME = 'f'

    p = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                description    =f'Print the contents of a file (binary / no encoding)\nThis tool is especially useful to find pesky carriage returns ({repr('\r')}).')
    p.add_argument(f'{A.FILENAME}',
                   help='file name')
    get = p.parse_args().__getattribute__
    # do it
    with open(get(A.FILENAME), 'rb') as f:

        print(f.read())

if __name__ == '__main__': main()



import socket
import datetime
import time

def do_it(ip_addr:str,
          port   :int,
          auto   :bool=False):

    port = int(port) if isinstance(port,str) else port
    s = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
    def send(msg:str):

        msg = (msg+'\n').encode()
        print(f'{datetime.datetime.now().isoformat()} :: {':'.join(map(str,s.getsockname()))}->{':'.join(map(str,s.getpeername()))} :: {msg}')
        s.sendall(msg)

    s.connect((ip_addr,port,))
    empty_count = 0
    if not auto:

        while True:

            if empty_count == 2: break
            if empty_count == 1:

                print('Enter an empty message again to break')

            msg = input(f'>>> ')
            if not msg:

                empty_count += 1
                continue

            empty_count = 0
            send(msg)
    
    else:

        msg = 'Hello, server!'
        print(f'auto = {auto}')
        while True:

            send(msg)
            time.sleep(1)

def main():

    import argparse

    class A:

        NET_ADDRESS = 'a'
        AUTO        = 'auto'

    p = argparse.ArgumentParser(description='Listen on a given network address and print to the console incoming data')
    p.add_argument(f'{A.NET_ADDRESS}',
                   help  =f'network address to talk to, in the form {{IP address}}:{{TCP port}}')
    p.add_argument(f'--{A.AUTO}',
                   help  =f'send messages automatically - useful if the intention is simply to observe at the server-side whether the server is receiving messages (and, thus, functioning)',
                   action='store_true')
    get          = p.parse_args().__getattribute__
    ip_addr,port = get(A.NET_ADDRESS).split(':')
    do_it(ip_addr=ip_addr,
          port   =port,
          auto   =get(A.AUTO))

if __name__ == '__main__': main()



import random

_BASE64_CHARS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-'

def get():

    return ''.join(random.choice(_BASE64_CHARS) for i in range(22))

def main():

    import argparse

    p = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                description    ='Get a base-64 identifier1\nUseful to generate Kafka cluster IDs, etc')
    p.parse_args()
    print(get())

if __name__ == '__main__': main()


import typing

def do_it(fn     :str,
          enc    :str   =None,
          no_tabs:bool=False):

    preproc:typing.Callable[[str],str] = (lambda content: content) if not no_tabs else \
                                         (lambda content: content.replace('\t', ' '))
    with open(fn, mode='r', encoding=enc if enc is not None else 'utf-8') as f:

        print(preproc(f.read()))

def main():

    import argparse

    class A:

        FILE_NAME = 'f'
        ENCODING  = 'enc'
        NO_TABS   = 'notab'

    p = argparse.ArgumentParser(description='Print the contents of a file')
    p.add_argument(f'{A.FILE_NAME}',
                   help='file name')
    p.add_argument(f'--{A.ENCODING}',
                   help='file encoding - default to \'utf-8\'')
    p.add_argument(f'--{A.NO_TABS}',
                   help='print tabs as simple spaces',
                   action='store_true')
    args = p.parse_args()
    def get(a:str): return getattr(args, a)
    # do it
    do_it(fn     =get(A.FILE_NAME),
          enc    =get(A.ENCODING),
          no_tabs=get(A.NO_TABS))

if __name__ == '__main__': main()



import os
import os.path
import re

def do_it(wd      :str,
          fn_regex:str):

    print(f'Looking in {wd}, where file names match {repr(fn_regex)}')
    found:list[str] = []
    for dp,dns,fns in os.walk(wd):

        for fn in fns:

            fn_full  = os.path.join(dp, fn)
            if not re.search(pattern=fn_regex, string=fn_full): continue
            found.append(fn_full)
    
    if found:

        for fn in found:
                
            print('>>> ' + fn)
    
    else:

        print('No occurrences.')

def main():

    import argparse

    class A:

        WORKING_DIRECTORY = 'wd'
        FILE_NAME_REGEX   = 'fnre'

    p = argparse.ArgumentParser(description='Find all files whose name matches a given expression')
    p.add_argument(f'--{A.WORKING_DIRECTORY}',
                   help   ='working directory - defaults to current',
                   default='.')
    p.add_argument(f'{A.FILE_NAME_REGEX}',
                   help   ='file name regex to consider')
    get = p.parse_args().__getattribute__
    # do it
    do_it(wd      =get(A.WORKING_DIRECTORY),
          fn_regex=get(A.FILE_NAME_REGEX))

if __name__ == '__main__': main()



import datetime
import socket
import sys
import time
import typing

TIMEOUT_MS_DEFAULT     = 50
PORT_RANGE_DELIMITER   = '-'
FOLLOW_STEP_MS_DEFAULT = 500

def test(ip_addr   :str,
         port      :int,
         bind      :bool=False,
         timeout_ms:int =TIMEOUT_MS_DEFAULT) -> bool:

    if not bind:

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(timeout_ms/1000)
        try:

            s.connect((ip_addr, port))
            
        except: return False
        else:
        
            s.close()
            return True
    
    else:

        s = socket.socket()
        try:

            s.bind((ip_addr, port))
        
        except: return False
        else:
        
            s.close()
            return True

def do_it(addr          :str,
          bind          :bool    =False,
          timeout_ms    :int     =TIMEOUT_MS_DEFAULT,
          follow        :bool    =False,
          follow_step_ms:int|None=None,
          follow_cb     :typing.Callable[[bool],None]=lambda up,addr: print(f'{addr} | {datetime.datetime.now().isoformat()} | {'OK!' if up else 'Not OK...'}')) -> bool:

    ip_addr,port = addr.split(':')
    if follow_step_ms is None: follow_step_ms = FOLLOW_STEP_MS_DEFAULT
    if follow:

        testf                = lambda: test(ip_addr=ip_addr, port=int(port), bind=bind, timeout_ms=timeout_ms)
        state_prev:bool|None = None
        try:

            while True:

                state = testf()
                if (state_prev is None) or (state != state_prev):

                    follow_cb(state, addr)

                state_prev = state
                time.sleep(follow_step_ms/1000)
        
        except KeyboardInterrupt as ex:

            print('Done')

    elif PORT_RANGE_DELIMITER in port:

        tested    = []
        tested_ok = []
        for port_ in map(str, range(*map(int, map(str.strip, port.split(PORT_RANGE_DELIMITER))))):

            tested.append(port_)
            print("Port " + port_, end=': ')
            if do_it(addr=f'{ip_addr}:{port_}',bind=bind,timeout_ms=timeout_ms):

                tested_ok.append(port_)
            
        print('OK ports:\n    {}'.format(', '.join(tested_ok)))
        return tested_ok

    else:
    
        ok = test(ip_addr=ip_addr, port=int(port), bind=bind, timeout_ms=timeout_ms)
        print("OK" if ok else "NOT OK AT ALL!!!")
        return ok

def main():

    import argparse

    class A:

        ADDRESS     = 'a'
        BIND        = 'bind'
        TIMEOUT     = 't'
        FOLLOW      = 'f'
        FOLLOW_STEP = 'fs'

    class DEFAULTS:

        TIMEOUT = TIMEOUT_MS_DEFAULT

    p = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                description    ='Test if a network address can be connected to / is hosting a TCP server')
    p.add_argument(f'{A.ADDRESS}',
                   help   =f'network address to test, in the form {{IP address}}:{{TCP port}}\nA range of ports to test may be given by delimiting the 1st and last ports with {repr(PORT_RANGE_DELIMITER)}')
    p.add_argument(f'--{A.BIND}',
                   help   =f'whether to test binding\nBy default, this tool tests that a connection to the given network address is possible. If this option is set, what is tested instead is whether the given network address is bindable i.e. can be listened on by a server.',
                   action ='store_true')
    p.add_argument(f'--{A.TIMEOUT}',
                   help   =f'timeout (in milliseconds) to consider in waiting for the connection\nDefaults to {repr(DEFAULTS.TIMEOUT)}',
                   type   =int,
                   default=DEFAULTS.TIMEOUT)
    p.add_argument(f'--{A.FOLLOW}',
                   help   =f'whether to follow the changes in the connection availability\nChanges are logged to the console',
                   action ='store_true')
    p.add_argument(f'--{A.FOLLOW_STEP}',
                   help   =f'step [milliseconds] with which to follow / to check connection availability\nDefault: {FOLLOW_STEP_MS_DEFAULT}\nIf this option is set, there is not need to give option --{repr(A.FOLLOW)}.',
                   type   =int)
    # parse
    get = p.parse_args().__getattribute__
    def get_or(a:str, dv): (lambda v: v if v is not None else dv)(get(a))
    # do it
    do_it(addr          =get   (A.ADDRESS),
          bind          =get   (A.BIND),
          timeout_ms    =get   (A.TIMEOUT),
          follow        =get   (A.FOLLOW) or get(A.FOLLOW_STEP),
          follow_step_ms=get_or(A.FOLLOW_STEP, FOLLOW_STEP_MS_DEFAULT))

if __name__ == '__main__': main()



import os.path
import shutil

def main(src:str,
         dst:str):
    
    shutil.copytree(src=src,dst=dst)

def main():

    import argparse
    
    class A:

        SOURCE      = 's'
        DESTINATION = 'd'

    p = argparse.ArgumentParser(description='Copy a directory')
    p.add_argument(f'{A.SOURCE}',
                   help='source directory')
    p.add_argument(f'{A.DESTINATION}',
                   help='destination directory')
    get = p.parse_args().__getattribute__
    main(src=get(A.SOURCE),
         dst=get(A.DESTINATION))

if __name__ == '__main__': main()


import abc
import typing

import paramiko

def _prompt_handler(title:str, instructions:str, prompts:list[tuple[str,bool]]):

    print(title)
    print(instructions)
    return tuple(input(f'{prompt[0]}: ') for prompt in prompts)

def get_paramiko(ip_addr :str,
                 user    :str,
                 pwd     :str|None,
                 interact:bool=False,
                 pkey    :paramiko.RSAKey|None=None):

    if not interact:

        client = paramiko.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect(hostname=ip_addr, username=user, password=pwd, allow_agent=False)

    else:

        client = paramiko.Transport(ip_addr)
        client.connect(username=user)
        client.auth_interactive(username=user, handler=_prompt_handler)

    return client

def repl(client    :paramiko.SSHClient,
         prompt    :typing.Callable[[str],str] =lambda dir: f'{dir} > ',
         command_cb:typing.Callable[[str],None]=lambda cmd: None):

    current_dir = client.exec_command('pwd')[1].read().decode().splitlines()[0]
    try:

        while True:

            command = input(prompt(current_dir))
            command = '; '.join(filter(bool, (f'cd "{current_dir}"',command,'pwd')))
            stdin,stdout,stderr = client.exec_command(command)
            out_lines,err_lines = stdout.read().decode().splitlines(), \
                                  stderr.read().decode().splitlines()
            current_dir         = out_lines.pop()
            output              = f'{'\n'.join(out_lines)}{'\n'.join(err_lines)}'
            if output: print(output)
            command_cb(command)

    except KeyboardInterrupt: pass

def main():

    import argparse

    class A: 
        
        IP_ADDRESS     = 'ip'
        USERNAME       = 'user'
        PASSWORD       = 'pass'
        AUTH_INTERACT  = 'inter'
        PROMPT_VERBOSE = 'v'

    p = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                description    ='SSH REPL\nTo quit, give an empty command twice.')
    p.add_argument(f'{A.IP_ADDRESS}',
                   help='SSH server IP address / host name')
    p.add_argument(f'--{A.USERNAME}',
                   help='user name')
    p.add_argument(f'--{A.PASSWORD}',
                   help='password')
    p.add_argument(f'--{A.AUTH_INTERACT}',
                   help  ='whether to request interactive authentication',
                   action='store_true')
    p.add_argument(f'--{A.PROMPT_VERBOSE}',
                   help  ='make the prompt more verbose - display host address and current directory',
                   action='store_true')
    def get(a:str,_args=p.parse_args()): return getattr(_args,a)
    # do it
    client = get_paramiko(ip_addr =get(A.IP_ADDRESS),
                          user    =get(A.USERNAME),
                          pwd     =get(A.PASSWORD),
                          interact=get(A.AUTH_INTERACT))
    try    : repl(client, prompt=(lambda dir: '> ') if not get(A.PROMPT_VERBOSE) else (lambda dir: f'{get(A.IP_ADDRESS)}{dir}\n> '))
    finally: 
        
        client.close()
        print('Done')

if __name__ == '__main__': main()



import datetime
import socketserver

class ConsoleRequestHandler(socketserver.StreamRequestHandler):

    def handle(self):

        while True:
            
            msg = self.rfile.readline()
            if not msg: break
            print(f'{datetime.datetime.now().isoformat()} :: {':'.join(map(str,self.request.getpeername()))}->{':'.join(map(str,self.request.getsockname()))} :: {msg}')

def do_it(ip_addr:str,
          port   :int):

    port = int(port) if isinstance(port,str) else port
    with socketserver.TCPServer((ip_addr,port,), ConsoleRequestHandler) as server:

        server.serve_forever()

def main():

    import argparse

    class A:

        NET_ADDRESS = 'a'

    p = argparse.ArgumentParser(description='Listen on a given network address and print to the console incoming data')
    p.add_argument(f'{A.NET_ADDRESS}',
                   help  =f'network address to listen on, in the form {{IP address}}:{{TCP port}}')
    get          = p.parse_args().__getattribute__
    ip_addr,port = get(A.NET_ADDRESS).split(':')
    do_it(ip_addr=ip_addr,port=port)

if __name__ == '__main__': main()



import os
import os.path
import re
import typing

from   jl95terceira.pytools import printf

def do_it(wd     :str,
          fnpat  :str                            =None,
          encf   :typing.Callable[[str,str], str]=None,
          no_tabs:bool                           =False):

    if encf  is None: encf  = lambda base,ext: 'utf-8'
    if fnpat is None: fnpat = '.*'
    fnre = re.compile(fnpat)
    print('')
    for dp,rdnn,rfnn in os.walk(top=wd):

        for rfn in rfnn:

            fn_full = os.path.join(dp,rfn)
            fn      = os.path.relpath(fn_full, start=wd)
            if not fnre.search(fn): continue
            print(f'{128*'-'} {fn}')
            try:
                
                printf.do_it(fn=fn_full,enc=encf(*os.path.splitext(fn)),no_tabs=no_tabs)
                print(f'{128*'-'}\n')
            
            except Exception as ex:

                print(f'error on printing file {fn} - {ex}')
            
def main():

    import argparse

    class A:

        WORKING_DIR       = 'wd'
        FILE_FILTER_REGEX = 'fre'
        ENCODING_FUNC     = 'encf'
        NO_TABS           = 'notab'

    p = argparse.ArgumentParser(description='Print the contents of all files in a given directory')
    p.add_argument(f'--{A.WORKING_DIR}',
                   help   ='working directory - defaults to current',
                   default='.')
    p.add_argument(f'--{A.FILE_FILTER_REGEX}',
                   help   ='regular expression to filter files by name to print - defaults to match any name (print all files)')
    p.add_argument(f'--{A.ENCODING_FUNC}',
                   help   ='file encoding function(base name, extension) literal - defaults to always return \'utf-8\'')
    p.add_argument(f'--{A.NO_TABS}',
                   help='print tabs as simple spaces',
                   action='store_true')
    args = p.parse_args()
    def get(a:str): return getattr(args, a)
    # do it
    do_it(wd     =     get(A.WORKING_DIR),
          fnpat  =     get(A.FILE_FILTER_REGEX),
          encf   =eval(get(A.ENCODING_FUNC)) if get(A.ENCODING_FUNC) is not None else None,
          no_tabs=     get(A.NO_TABS))

if __name__ == '__main__': main()



import os
import os.path
import re
import typing
import paramiko

from jl95terceira.batteries import *

def do_it(path_src    :str,
          fn_filter   :typing.Callable[[str],bool],
          ssh_ip_addr :str,
          ssh_port    :int,
          ssh_username:str,
          ssh_password:str,
          path_dst    :str):
    
    c = paramiko.SSHClient()
    c.set_missing_host_key_policy(paramiko.AutoAddPolicy)
    c.connect(hostname=ssh_ip_addr, port=ssh_port, username=ssh_username, password=ssh_password)
    fc = c.open_sftp()
    if os.path.isdir(path_src):

        for dp,rdnn,rfnn in os.walk(path_src):

            for rfn in rfnn:

                rfn_src = os.path.relpath(path=os.path.join(dp, rfn), start=path_src)
                if not fn_filter(rfn_src): continue
                fn_src  = os.path.join(path_src,rfn_src)
                pad     = ' '*(max((0,40-len(fn_src),)))
                fn_dst  = '/'.join(os.path.join(path_dst, rfn_src).split(os.path.sep))
                print(f'Copy from local {repr(fn_src)}{(pad)} to remote {repr(fn_dst)}', end='')
                fn_dst_tokens = fn_dst.split('/')
                for i in range(len(fn_dst_tokens)-1):

                    try: fc.mkdir('/'.join(fn_dst_tokens[:1+i]))
                    except: pass

                try:
                
                    fc.put(localpath=fn_src, remotepath=fn_dst)

                except Exception as e: print(f'{pad} - {e}')
                else:                  print('')
            
    else:

        pass

def main():

    import argparse

    class A:

        SOURCE_PATH         = 'spath'
        FILENAME_REGEX      = 'fre'
        DESTINATION_ADDRESS = 'daddr'
        DESTINATION_USER    = 'duser'
        DESTINATION_PASS    = 'dpass'
        DESTINATION_PATH    = 'dpath'

    p = argparse.ArgumentParser(description='Copy files to a remote directory via FTP / SSH')
    p.add_argument(f'{A.SOURCE_PATH}',
                   help='source (local) path')
    p.add_argument(f'--{A.FILENAME_REGEX}',
                   help='source (local) file name regex, to filter files to copy')
    p.add_argument(f'{A.DESTINATION_ADDRESS}',
                   help='destination (remote) SSH address in form \'{ip address}:{port}\'')
    p.add_argument(f'{A.DESTINATION_USER}',
                   help='destination (remote) user name')
    p.add_argument(f'{A.DESTINATION_PASS}',
                   help='destination (remote) password')
    p.add_argument(f'{A.DESTINATION_PATH}',
                   help='destination (remote) path')
    get = p.parse_args().__getattribute__
    # do it
    ip_addr,port = (lambda a,b: (a, int(b)))(*map(str.strip, str(get(A.DESTINATION_ADDRESS)).split(':')))
    print([ip_addr, port])
    do_it(path_src    =get(A.SOURCE_PATH),
          fn_filter   =re.compile(pattern=get(A.FILENAME_REGEX)).search if get(A.FILENAME_REGEX) is not None else constant(True),
          ssh_ip_addr =ip_addr,
          ssh_port    =port,
          ssh_username=get(A.DESTINATION_USER),
          ssh_password=get(A.DESTINATION_PASS),
          path_dst    =get(A.DESTINATION_PATH))

if __name__ == '__main__': main()



import os
import os.path
import subprocess
import typing

def do_it(fn       :str,
          wd       :str|None  =None,
          ow_prompt:typing.Callable[[],bool]|None=None,
          c        :bytes|None=None):

    if ow_prompt is None: ow_prompt = lambda: False
    fp = os.path.join(wd, fn) if wd is not None else \
         fn
    if os.path.exists(fp):

        print(f'A file already exists at {repr(fp)}')
        if not ow_prompt():
            
            return False

    with open(fp, mode='wb') as f: 

        if c is not None: f.write(c)
        else: pass

    return True

def main():

    import argparse
    from   jl95terceira.pytools.envlib.vars.builtin import EDITOR

    class A:

        WORKING_DIR     = 'wd'
        FILE_NAME       = 'f'
        FORCE_OVERWRITE = 'ow'
        CONTENT         = 'c'
        OPEN            = 'o'
    
    p = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                description    ='Create a file')
    p.add_argument(f'{A.FILE_NAME}',
                   help='file name')
    p.add_argument(f'--{A.WORKING_DIR}',
                   help='directory at which to create the file\nDefaults to current',
                   default='.')
    p.add_argument(f'--{A.FORCE_OVERWRITE}',
                   help='force ovewrite the file, if it already exists - for if you like DANGER',
                   action='store_true')
    p.add_argument(f'--{A.CONTENT}',
                   help='content with which to initialize / to write to the file\nDefaults to none i.e. create an empty file')
    p.add_argument(f'--{A.OPEN}',
                   help='open the file with the default editor, after creating the file',
                   action='store_true')
    # do it
    get = p.parse_args().__getattribute__
    OVERWRITE_CONFIRMATION_WORD = 'ow'
    success = do_it(fn       =get(A.FILE_NAME),
                    wd       =get(A.WORKING_DIR),
                    ow_prompt=(lambda: True) if get(A.FORCE_OVERWRITE) else (lambda: (input(f'Enter {repr(OVERWRITE_CONFIRMATION_WORD)}, to overwrite: ') == OVERWRITE_CONFIRMATION_WORD)),
                    c        =get(A.CONTENT))
    
    if success and get(A.OPEN):

        os.system(EDITOR.get()(get(A.FILE_NAME)))

if __name__ == '__main__': main()



import os
import os.path

from   jl95terceira         import batteries
from   jl95terceira.pytools import envlib as env

THIS_DIR         = os.path.split(__file__)[0]
OUTPUT_FILE_PATH = os.path.join(batteries.os.TEMP_DIR,'pytools-wiki.md')

def do_it(wikifn:str):

    verbose_original       = env.state.State.verbos
    env.state.State.verbos = env.state.Verbosities.OFF
    def get_fns():

        return os.listdir(THIS_DIR)

    pad = max(map(len,get_fns()))
    with open(wikifn, 'w', encoding='utf-8') as f:

        f.write('\n'.join((

            'This wiki is generated using <code>wikimaker</code>.',
            '',
            
        )))
        f.write('<ul>')
        for fn in get_fns():

            fn_full = os.path.join(THIS_DIR,fn)
            if not batteries.is_module(fn_full): continue
            print(fn,end='')
            stdout = os.popen(f'python {os.path.join(THIS_DIR,fn)} -h').read()
            if not stdout.strip(): 
                
                print(f'{' '*(pad-len(fn))} - no CLI')
                continue

            f.write(f'\n\n<li>')
            f.write(f'\n\n<b>{os.path.splitext(fn)[0]}</b>')
            f.write(f'\n\n```\n{'\n'.join(stdout.splitlines())}\n```\n')
            f.write(f'\n\n</li>')
            print('')
        
        f.write('</ul>\n\n')
    
    print(f'Done.\nWritten to: {wikifn}')
    env.state.State.verbos = verbose_original

def main():

    import argparse

    p = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                description    ='A helping tool to remake the wiki')
    p.add_argument('--f',
                   help   =f'file to which to write the wiki - default to {repr(OUTPUT_FILE_PATH)}',
                   default=OUTPUT_FILE_PATH)
    args = p.parse_args()
    do_it(wikifn=args.f)

if __name__ == '__main__': main()



import dataclasses
import mmap
import typing

@dataclasses.dataclass
class Result:

    section:str
    begin  :int
    end    :int

def do_it(fn1 :str,
          fn2 :str):
       
    with open(fn1, mode='rb') as f1:
            
        m1 = mmap.mmap(f1.fileno(), 0, access=mmap.ACCESS_READ)
        with open(fn2, mode='rb')  as f2:
            
            m2 = mmap.mmap(f2.fileno(), 0, access=mmap.ACCESS_READ)
            # do it
            if len(m1) == len(m2): return Result(section='', begin=0, end=0)
            raise NotImplementedError()

def main():
    
    import argparse

    class A:
        
        FILE_1 = 'fn1'
        FILE_2 = 'fn2'
        ENCODING_1 = 'enc1'
        ENCODING_2 = 'enc2'

    p = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                description    ='Find the difference between two files')
    p.add_argument(f'{A.FILE_1}',
                   help='file #1')
    p.add_argument(f'{A.FILE_2}',
                   help='file #2')
    get = p.parse_args().__getattribute__
    # do it
    res = do_it(fn1 =get(A.FILE_1),
                fn2 =get(A.FILE_2))
    if not res.section:

        print('Files are equal')

    else:

        print('\n\n'.join((

            f'Lines {res.begin} to {res.end}:',
            res.section,

        )))

if __name__ == '__main__': main()



import os
import os.path
import re
import typing

def _do_it(wd     :str,
           handler:typing.Callable[[str],None],
           _start :str|None=None):

    if _start is None:

        _start = wd

    for name in os.listdir(wd):

        name_full = os.path.join(wd, name)
        if os.path.isdir(name_full):
        
            name_rel = os.path.relpath(name_full, start=_start)
            handler(name_rel)
            _do_it(wd=name_full,handler=handler,_start=_start)

def do_it(wd    :str,
          fn_out:str=None,
          filter:typing.Callable[[str],bool]=lambda s: True):
    
    def _filtered(f:typing.Callable[[str],None]):

        def f_actual(s:str):

            if not filter(s): return
            f(s)
        
        return f_actual

    if fn_out is None: 

        @_filtered
        def _handler(s:str):

            print(s)

        _do_it(wd=wd, handler=_handler)

    else:

        with open(fn_out, mode='w', encoding='utf-8') as f:

            @_filtered
            def _handler(s:str):

                print(s)
                f.write(s+'\n')

            _do_it(wd=wd, handler=_handler)

def main():

    import argparse

    class A:

        WORKING_DIR  = 'wd'
        DUMP_FILE    = 'o'
        REGEX_FILTER = 'fre'

    p = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                description='List directories')
    p.add_argument(f'--{A.WORKING_DIR}',
                   help='working directory\nDefaults to current',
                   default='.')
    p.add_argument(f'--{A.DUMP_FILE}',
                   help='file to which to dump the analysis, if desired')
    p.add_argument(f'--{A.REGEX_FILTER}',
                   help=f'regular expression to consider as filter for directory names\nDefaults to exclude files starting with {'.'} (such as .git)',
                   default=r'^(?!\.git)')
    p.parse_args()
    def get(a:str,_args=p.parse_args()): return getattr(_args,a)
    # do it
    do_it(wd    =get(A.WORKING_DIR),
          fn_out=get(A.DUMP_FILE),
          filter=re.compile(get(A.REGEX_FILTER)).search)
    
    if __name__ == '__main__': main()



import datetime
import operator
import os
import os.path

from   jl95terceira.pytools import hashf

def do_it(h    :str,
          wd   :str,
          depth:int,
          fn   :str):

    hvv_list = hashf.Hasher(hasht=hashf.map_by_names()[h]).of_dir_by_name_to_map(root=wd,depth_filter=(lambda d: d <= depth) if depth is not None else (lambda d: True))
    pad      = max(map(len,map(operator.itemgetter(0),hvv_list)))
    t_now    = datetime.datetime.now().isoformat()
    log      = '\n'.join((

        'Directory:{}'.format(os.path.abspath(wd)),
        'Time     :{}'.format(t_now),
        'Hashes   :',
        *('    {}{} = {}'.format(n,' '*(pad-len(n)),hv) for n,hv in sorted(hvv_list))

    ))+'\n'
    print(log)
    if fn is not None:

        with open(fn, mode='a') as f:

            f.write(log)

def main():

    import argparse

    class A:

        HASH_FUNCTION     = 'h'
        WORKING_DIRECTORY = 'wd'
        DEPTH             = 'depth'
        OUTPUT_FILE       = 'o'

    class DEFAULTS:

        HASH_FUNCTION = 'sha256'

    p = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                description    ='Hash directories recursively and print the results')
    p.add_argument(f'--{A.HASH_FUNCTION}',
                   help=f'hash function name\nDefaults to {DEFAULTS.HASH_FUNCTION}\nPossible values: {hashf.names()}',
                   default=DEFAULTS.HASH_FUNCTION)
    p.add_argument(f'--{A.WORKING_DIRECTORY}',
                   help='working directory - default to current',
                   default='.')
    p.add_argument(f'--{A.DEPTH}',
                   help='directory depth limit for hashing, if desired')
    p.add_argument(f'--{A.OUTPUT_FILE}',
                   help='name of file to which to write results, if desired')
    def get(a:str,_args=p.parse_args()): return getattr(_args,a)
    do_it(h    =get(A.HASH_FUNCTION),
          wd   =get(A.WORKING_DIRECTORY),
          depth=int(get(A.DEPTH))      if get(A.DEPTH)       else None,
          fn   =    get(A.OUTPUT_FILE) if get(A.OUTPUT_FILE) else None)
    
if __name__ == '__main__': main()



import hashlib
import os
import os.path
import typing

def _path_partials(path:str):

    return path.split(os.path.sep)

class _HashFactory:

    def __call__(self, *a) -> '_Hash': pass

class _Hash: # dummy class for type hinting purposes to match "hashlib._hashlib.HASH" which, for a weird reason, is recognized by vscode as "hashlib._Hash" - hence the class' name

    def update   (self, bb): pass
    def hexdigest(self) -> str: return ""

class Hasher:

    def __init__(self,hasht:_HashFactory):

        self.hasht = hasht

    def of_file_by_name      (self, fn  :str) -> str:

        with open(fn,'rb') as fd:

            hv = self.of_file(fd)
            return hv

    def of_file              (self, f   :typing.BinaryIO) -> str:

        return self.hasht(f.read()).hexdigest()

    def of_dir_by_name       (self, root:str,
                             _start     :str=None,
                             _hvcb      :typing.Callable[[str,bool,str],None]=(lambda *a,**ka: None),
                             _depth     :int=-1) -> str:
        
        _start = _start if _start is not None else root
        h = self.hasht()
        _depth += 1
        for n in sorted(os.listdir(root)):

            joinedpath = os.path.join(root,n)
            relpath    = os.path.relpath(path=joinedpath,start=_start)
            for p in _path_partials(relpath):

                h.update(p.encode())

            hv = self.of_file_by_name(fn    =joinedpath) if not os.path.isdir(joinedpath) else \
                 self.of_dir_by_name (root  =joinedpath,
                                             _start=_start, 
                                             _hvcb =_hvcb, 
                                             _depth=_depth)
            _hvcb(relpath, _depth, hv)
            h.update(hv.encode())

        return h.hexdigest()

    def of_dir_by_name_to_map(self,
                              root        :str,
                              depth_filter:typing.Callable[[int],bool]=lambda depth: depth < 1):

        hvv_list:list[tuple[str,str]] = list()
        def cb(relpath,depth,hv):

            if not depth_filter(depth): return
            hvv_list.append((relpath, hv))

        hvv_list.append(('.', self.of_dir_by_name(root=root,_hvcb=cb,)))
        return hvv_list
        
    def of                   (self, path:str, 
                              _start:str=None):

        return self.of_file_by_name(path) if os.path.isfile(path) else \
               self.of_dir_by_name (root=path,_start=_start if _start is not None else \
                                                     path)

_PREDEFINED_HASHF_FACTORIES_MAP:dict[str,_HashFactory] = {
    
    'md5'   : hashlib.md5,
    'sha1'  : hashlib.sha1,
    'sha224': hashlib.sha224,
    'sha256': hashlib.sha256,
    'sha384': hashlib.sha384,
    'sha512': hashlib.sha512
}

def map_by_names(): return _PREDEFINED_HASHF_FACTORIES_MAP

_NAMES = tuple(_PREDEFINED_HASHF_FACTORIES_MAP)

def names(): return _NAMES

def _asserted_hashf(h:str):

    if h not in _PREDEFINED_HASHF_FACTORIES_MAP:

        raise Exception(f"unknown hash type {repr(h)}\navailable: {_PREDEFINED_HASHF_FACTORIES_MAP.keys()}")

    return _PREDEFINED_HASHF_FACTORIES_MAP[h]
    
def main():
    
    import argparse

    class A:

        HASH_FUNCTION = 'h'
        PATH          = 'path'
        LIST          = 'list'
        DEPTH_MAX     = 'depth'

    class DEFAULTS:

        HASH_FUNCTION = 'sha256'
        DEPTH         = 'inf'

    p = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                description    ='Hash a file or a directory')
    p.add_argument(f'--{A.HASH_FUNCTION}',
                   help=f'hash function name\nPossible values: {list(_PREDEFINED_HASHF_FACTORIES_MAP)}\nDefaults to {repr(DEFAULTS.HASH_FUNCTION)}.',
                   default=DEFAULTS.HASH_FUNCTION)
    p.add_argument(f'{A.PATH}',
                   help='path of file or of directory to hash')
    p.add_argument(f'--{A.LIST}',
                   help='if hashing a directory: list the hashes of all files and directories within',
                   action='store_true')
    def _depth(a):

        if a is not None:
            try: return int(a)
            except ValueError:
                try:
                    if str(a).lower() == DEFAULTS.DEPTH: return float('inf')
                except: raise Exception(f'not a valid depth value - must be integer or {repr(DEFAULTS.DEPTH)}: {a}')

    p.add_argument(f'--{A.DEPTH_MAX}',
                   help   =f'if hashing a directory: specify for which depth at most to display the hashes of the files / directories at such depths - numeric or {repr(DEFAULTS.DEPTH)}\nDefaults to 1',
                   type   =_depth,
                   default=1)
    # parse
    args = p.parse_args()
    def get(a:str): return getattr(args, a)
    # do it
    if not get(A.LIST):
        
        print(Hasher(hasht=_asserted_hashf(get(A.HASH_FUNCTION))).of(get(A.PATH)))
    
    else:

        hashes_list = Hasher(hasht=_asserted_hashf(get(A.HASH_FUNCTION))).of_dir_by_name_to_map(root        =get(A.PATH), 
                                                                                                depth_filter=lambda depth: depth < get(A.DEPTH_MAX))
        padl        = max(map(len,map(lambda t: t[0], hashes_list)))
        print('\n'.join(f'{file_name}{(padl - len(file_name))*' '} ---- {hash_value}' for file_name,hash_value in hashes_list))

if __name__ == '__main__': main()



import ssl

import websockets.sync.client

def do_it(url     :str,
          use_ssl :bool=False,
          auto_ssl:bool=False,
          protocol:ssl._SSLMethod=None):

    if auto_ssl:
        
        use_ssl = url.startswith('wss')

    with websockets.sync.client.connect(url)                                                                                                 if not use_ssl else \
         websockets.sync.client.connect(url,ssl_context=ssl.SSLContext(protocol=protocol if protocol is not None else ssl.PROTOCOL_TLSv1_2)) as websocket:

        while (True):
            
            message = websocket.recv()
            print(f"Received: {message}")

def main():

    import argparse

    SSL_PROTOCOL_VERSION_DICT = {

        **{v:ssl.PROTOCOL_TLSv1 for v in ('1',
                                          '1.0')},
        '1.1' :ssl.PROTOCOL_TLSv1_1,
        '1.2' :ssl.PROTOCOL_TLSv1_2,
    }
    SSL_PROTOCOL_VERSION_DEFAULT = object()
    
    class A: 

        NET_ADDRESS   = 'addr'
        ASSUME_SSL    = 'ssl'
        ASSUME_NO_SSL = 'nossl'

    p = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                description    ='Connect to a websocket server')
    p.add_argument(f'{A.NET_ADDRESS}', 
                   help='websocket server network address')
    p.add_argument(f'--{A.ASSUME_SSL}', 
                   help   ='force assume SSL and, if a value is given, assume the value as the SSL protocol version (default = \'1.2\')\nThis option is mutually exclusive with option \'--nossl\'.',
                   nargs  ='?',
                   default=None,
                   const  =SSL_PROTOCOL_VERSION_DEFAULT)
    p.add_argument(f'--{A.ASSUME_NO_SSL}',
                   help  ='force assume no SSL\nThis option is mutually exclusive with option \'--ssl\'.',
                   action='store_true')
    get = p.parse_args().__getattribute__
    if get(A.ASSUME_SSL) is not None and get(A.ASSUME_NO_SSL):

        emsg = f'options --{A.ASSUME_SSL} and --{A.ASSUME_NO_SSL} are mutually exclusive'
        raise Exception(emsg)

    do_it(url     =get(A.NET_ADDRESS),
          protocol=None if get(A.ASSUME_SSL) is None else SSL_PROTOCOL_VERSION_DICT[get(A.ASSUME_SSL)] if get(A.ASSUME_SSL) is not SSL_PROTOCOL_VERSION_DEFAULT else None,
          auto_ssl=get(A.ASSUME_SSL) is     None and not get(A.ASSUME_NO_SSL),
          use_ssl =get(A.ASSUME_SSL) is not None or  not get(A.ASSUME_NO_SSL))

if __name__ == '__main__': main()



import os.path
import shutil

def do_it(fpath_from:str,
         dir_to    :str,
         fname_to  :str):
    
    shutil.copyfile(src=fpath_from,
                    dst=os.path.join(dir_to, fname_to if fname_to else (os.path.split(fpath_from)[-1])))

def main():

    import argparse
    
    class A:

        SOURCE_PATH    = 'source'
        DEST_DIR       = 'destdir'
        DEST_FILE_NAME = 'name'

    p = argparse.ArgumentParser(description='Copy a file')
    p.add_argument(f'{A.SOURCE_PATH}',
                   help='source file path (full)')
    p.add_argument(f'{A.DEST_DIR}',
                   help='destination directory')
    p.add_argument(f'--{A.DEST_FILE_NAME}',
                   help='destination file name, if to be different')
    get = p.parse_args().__getattribute__
    do_it(fpath_from=get(A.SOURCE_PATH),
          dir_to    =get(A.DEST_DIR),
          fname_to  =get(A.DEST_FILE_NAME))

if __name__ == '__main__': main()



import dataclasses
import os
import os.path
import re
import typing

@dataclasses.dataclass
class Result:

    fn     :str
    n      :int
    matches:list[typing.Any]

def do_it(wd          :str,
          fn_regex    :str,
          encf        :typing.Callable[[str],str],
          string      :str,
          as_regex    :bool=False,
          show_matches:bool=True,
          no_errors   :bool=False):

    if not isinstance(encf, typing.Callable): raise Exception(f'encoding is expected as function of filename (callable) - got type {type(encf)} instead')
    pattern = re.compile(fn_regex)
    print(f'Looking for {repr(string)} in {wd}, where file names match {repr(fn_regex)}')
    results:list[Result] = []
    for full,rel in ((                os.path.join(dp, fn),
                      os.path.relpath(os.path.join(dp, fn), start=wd)) for dp,dns,fns in os.walk(wd) for fn in fns):

        if not pattern.search(rel): continue
        try:

            encoding = encf(full)
            with open(full, 'r', encoding=encoding) as f:

                finds = list()
                for line in f.readlines():

                    finds.extend(re.findall(pattern=f'{re.escape(string)if not as_regex else string}', string=line))
                
                if finds:
                    
                    results.append(Result(fn=full,n=len(finds),matches=finds))
        
        except Exception as e:

            if not no_errors:
                
                print(f'Exception at file {rel}: {e}')
    
    lens = list(map(len,(result.fn for result in results)))
    if lens:

        fnpad = max(lens)
        opad  = max(map(len,map(str,(result.n for result in results))))
        for result in results:

            print('>>> ' + result.fn + (fnpad-len(result.fn))*' ' + f' -> {(opad-len(str(result.n)))*' '}{result.n} occurrences{'' if not as_regex or not show_matches else f': {repr(sorted(set(match if isinstance(match,str) else match[0] for match in result.matches)))}'}')
    
    else:

        print('Not any file with occurrences.')

def main():

    import argparse

    class A:

        WORKIND_DIR       = 'wd'
        FILENAME_REGEX    = 'fre'
        ALL_FILES         = 'all'
        ENCODING          = 'enc'
        ENCODING_FUNCTION = 'encf'
        STRING            = 'string'
        STRING_LITERAL    = 'literal'
        REGEX             = 're'
        LESS              = 'less'
        NO_ERRORS         = 'noerror'

    p = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                description='Find a string within files\nFiles may be filtered by their names according to a given expression.\nThe string to search may be given as an expression.')
    p.add_argument(f'--{A.WORKIND_DIR}',
                   help   ='working directory',
                   default='.')
    p.add_argument(f'--{A.FILENAME_REGEX}',
                   help=f'file name regex to use for filtering files in which to find the given string\nDefaults to all files that don\'t start with a dot (".").\nYou can override this option to look in all files by giving option {repr(A.ALL_FILES)}.',
                   default='^(?!\\.)')
    p.add_argument(f'--{A.ALL_FILES}',
                   help=f'look in all files - equivalent to setting option {repr(A.FILENAME_REGEX)} as ".*"',
                   action='store_true')
    p.add_argument(f'--{A.ENCODING}',
                   help   ='the text encoding to consider for all files - defaults to \'utf-8\'')
    p.add_argument(f'--{A.ENCODING_FUNCTION}',
                   help   ='the text encoding to consider for each file, given the file name, as a function literal (\'lambda fn: ...\') - alternative to option \'--enc\'')
    p.add_argument(f'{A.STRING}',
                   help='string to look for in files')
    p.add_argument(f'--{A.STRING_LITERAL}',
                   help=f'take the given string as a Python string literal\nThis option allows for passing newlines as {repr('\n')}, tabs as {repr('\t')}, etc - literal backslashes ({repr('\\')}) must be doubled',
                   action='store_true')
    p.add_argument(f'--{A.REGEX}',
                   help='consider string as regex',
                   action='store_true')
    p.add_argument(f'--{A.LESS}',
                   help=f'do NOT print matches, in the case of regex search (less verbose)\nBy default, when matching with a regex via option {A.REGEX}, the matches are printed along with the occurrences.',
                   action='store_true')
    p.add_argument(f'--{A.NO_ERRORS}',
                   help=f'silence errors when they happen\nBy default, errors on opening files for reading are printed.',
                   action='store_true')
    get = p.parse_args().__getattribute__
    # do it
    enc_excl  = (A.ENCODING, A.ENCODING_FUNCTION,)
    if len(tuple(filter(lambda a: get(a) is not None, enc_excl))) > 1:

        raise Exception(f'all of options {enc_excl} given - only 1 allowed')

    do_it(wd          =get(A.WORKIND_DIR),
          fn_regex    =get(A.FILENAME_REGEX) if not get(A.ALL_FILES) else '.*',
          encf        =eval(get(A.ENCODING_FUNCTION)) if get(A.ENCODING_FUNCTION) is not None else (lambda fn,_enc=get(A.ENCODING): _enc) if get(A.ENCODING) is not None else lambda fn: 'utf-8',
          string      =get(A.STRING) if not get(A.STRING_LITERAL) else eval(get(A.STRING)),
          as_regex    =get(A.REGEX),
          show_matches=not get(A.LESS),
          no_errors   =get(A.NO_ERRORS))

if __name__ == '__main__': main()



import uuid

def get(upper :bool=False,
        squash:bool=False):

    return (lambda u: u.upper  ()       if upper  else u)\
          ((lambda u: u.replace('-','') if squash else u)\
           (str(uuid.uuid4())))

def main():

    import argparse

    class A:

        UPPER_CASE = 'u'
        SQUASH     = 's'

    p = argparse.ArgumentParser(description='Print a new UUID',
                                formatter_class=argparse.RawTextHelpFormatter)
    p.add_argument(f'--{A.UPPER_CASE}',
                   help='uppercase',
                   action='store_true')
    p.add_argument(f'--{A.SQUASH}',
                   help='remove hyphens',
                   action='store_true')
    # parse args
    get_ = p.parse_args().__getattribute__
    # do it
    print(get(upper =get_(A.UPPER_CASE),
              squash=get_(A.SQUASH)))

if __name__ == '__main__': main()
